import argparse
import glob
import logging
import os
import sys
from pathlib import Path
from pycapo import CapoConfig

from carta_envoy.launchers import CartaLauncher

"""
Setup and Launch CARTA via Workspaces
(modified from carta-valet)
"""

logger = logging.getLogger("carta_envoy")
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))


def _get_settings(location: str) -> dict:
    carta_settings = CapoConfig().settings("edu.nrao.carta")
    ws_carta = CapoConfig().settings("edu.nrao.workspaces.carta")
    return {
        "timeout": carta_settings.timeoutInMinutes,
        "carta_path": carta_settings.pathToExecutable,
        "redis_host": carta_settings.redisHost,
        "redis_port": carta_settings.redisPort,
        "redis_password": carta_settings.redisPassword,
        "reverse_proxy": carta_settings.reverseProxyHost,
        "data_location": Path(location),
        "useCarta": ws_carta.useCarta,
    }


def ensure_correct_directory(settings: dict):
    directory = settings["data_location"]
    logger.info(f"Current working directory: {os.getcwd()}")
    if os.getcwd() != Path(directory):
        os.chdir(directory)
    if len(glob.glob("*.fits")) != 0:
        logger.info(f"FITS files found in location '{directory}', continuing...")
        return
    else:
        logger.info(f"No FITS files found in location '{directory}', checking other directories...")

        contained_dirs = glob.glob("*/", recursive=True)
        for dir in contained_dirs:
            if len(glob.glob(dir + "*.fits")) != 0:
                path = Path(dir)
                os.chdir(path)
                logger.info(f"Correct data location found: {path}. Changing location to match...")


def make_arg_parser() -> argparse.ArgumentParser:
    """
    Create command-line argument parser

    :return: Argument parser
    """
    parser = argparse.ArgumentParser(
        description="Launch CARTA via Workspaces",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "-w",
        "--where",
        action="store",
        type=str,
        required=True,
        help="where to run CARTA",
    )
    return parser


def main():
    args = make_arg_parser().parse_args()
    settings = _get_settings(args.where)
    ensure_correct_directory(settings)
    carta = CartaLauncher(settings)
    carta.launch()
    carta.teardown()
    logger.info("CARTA session has completed.")
