from typing import Dict
from .utilities import generate_random_str

import logging
import redis
import socket

"""
Handle Redis server connection setup
"""


class RedisConnect:
    """
    class for managing connecting to Redis server
    """

    def __init__(self, settings: dict):
        self.logger = logging.getLogger("carta_envoy")
        self.settings = settings
        self.generated_ids = self.generate_ids()
        self.redis_values = None
        self.conn = None

    @staticmethod
    def generate_ids() -> dict:
        return {
            "front_end_id": generate_random_str(),
            "back_end_id": generate_random_str(),
            "system_name": generate_random_str(),
        }

    def generate_url(self) -> str:
        self.logger.info("Generating CARTA url...")
        front_end_id = self.generated_ids["front_end_id"]
        back_end_id = self.generated_ids["back_end_id"]
        proxy = self.settings["reverse_proxy"]
        carta_url = f"https://{proxy}/{front_end_id}/?socketUrl=wss://{proxy}/{back_end_id}/"
        return carta_url

    def get_available_port(self) -> int:
        """
        Asks the operating system for an available port on localhost using the socket module

        :return: Number of open port
        """
        self.logger.info("Finding available port...")
        with socket.socket(family=socket.AF_INET, type=socket.SOCK_STREAM) as s:
            # Bind socket to available port
            s.bind(("", 0))
            return s.getsockname()[1]

    def prepare_redis(self) -> dict:
        self.logger.info("Preparing Redis server...")
        front_end_port = self.get_available_port()
        back_end_port = self.get_available_port()
        carta_url = self.generate_url()
        self.conn = self.connect_to_redis()
        redis_values = self.get_redis_values(
            self.settings["reverse_proxy"],
            self.generated_ids["front_end_id"],
            self.generated_ids["back_end_id"],
            front_end_port,
            back_end_port,
        )
        self.set_redis_values(redis_values, int(self.settings["timeout"]))
        self.send_carta_url_to_rh(carta_url)

        return {
            "front_end_port": front_end_port,
            "back_end_port": back_end_port,
            "carta_url": carta_url,
        }

    def send_carta_url_to_rh(self, url: str):
        # TODO: implement sending url as webhook, WS-619
        pass

    def connect_to_redis(self) -> redis.Redis:
        """
        Connect to Redis server located at host:port

        :return: Successful connection to Redis server
        """
        self.logger.info("Connecting to Redis...")
        host = self.settings["redis_host"]
        port = self.settings["redis_port"]
        password = self.settings["redis_password"]

        conn = redis.Redis(host, port, password=password)
        # Ping the network to ensure the connection is sound
        conn.ping()

        return conn

    def get_redis_values(
        self,
        reverse_proxy_host: str,
        front_end_id: str,
        back_end_id: str,
        front_end_port: int,
        back_end_port: int,
    ) -> Dict[str, str]:
        """
        Get Redis key/value pairs for CARTA front end and back end setup

        :param reverse_proxy_host: Hostname for the reverse proxy
        :param front_end_id: Random string ID for CARTA front end
        :param back_end_id: Random string ID for CARTA back end
        :param front_end_port: Port to run the CARTA front end with
        :param back_end_port: Port to run the CARTA back end with
        :return: Dictionary of values
        """
        self.logger.info("Determining values for Redis...")
        service_name = generate_random_str()
        front_end = f"carta-front-{service_name}"
        back_end = f"carta-back-{service_name}"
        hostname = socket.getfqdn()
        values = {
            # FRONT END
            f"traefik/http/routers/{front_end}/rule": f"Host(`{reverse_proxy_host}`) && PathPrefix(`/{front_end_id}/`)",
            f"traefik/http/routers/{front_end}/service": front_end,
            f"traefik/http/services/{front_end}/loadbalancer/servers/0/url": f"http://{hostname}:{front_end_port}/",
            f"traefik/http/routers/{front_end}/middlewares/0": "stripPrefixFE@file",
            # BACK END
            f"traefik/http/routers/{back_end}/rule": f"Host(`{reverse_proxy_host}`) && PathPrefix(`/{back_end_id}/`)",
            f"traefik/http/routers/{back_end}/service": back_end,
            f"traefik/http/services/{back_end}/loadbalancer/servers/0/url": f"http://{hostname}:{back_end_port}/",
            f"traefik/http/routers/{back_end}/middlewares/0": "upgradeWSHeader@file",
            f"traefik/http/routers/{back_end}/middlewares/1": "stripPrefixFE@file",
        }

        unique_values = self.check_for_duplicate_values(values, front_end_port, back_end_port)
        self.redis_values = unique_values
        return unique_values

    def check_for_duplicate_values(self, redis_values: dict, front_port: int, back_port: int):
        self.logger.info("Checking for duplicate values on server...")
        for key in redis_values:
            if self.conn.get(key):
                self.logger.warning(
                    "WARNING: Redis value collision found. Generating new random IDs."
                )
                self.generated_ids = self.generate_ids()
                new_values = self.get_redis_values(
                    self.settings["reverse_proxy"],
                    self.generated_ids["front_end_id"],
                    self.generated_ids["back_end_id"],
                    front_port,
                    back_port,
                )
                return new_values

        return redis_values

    def set_redis_values(self, redis_values: dict, timeout_minutes: int):
        """
        Set Redis values with an expiration of 24 hours

        :param redis_values: Dictionary of Redis entries
        :param timeout_minutes: Timeout of the valet in minutes
        """
        self.logger.info(f"Setting Redis values with a timeout of {timeout_minutes} minutes...")
        for key, val in redis_values.items():
            self.conn.setex(key, 60 * timeout_minutes + 60, val)

    def remove_redis_values(self):
        self.logger.info("Removing Redis values for this CARTA session...")
        for key in self.redis_values:
            self.conn.delete(key)


class ArchiveConnect:
    """
    Class for updating AAT with information from CARTA envoy
    """

    def update_rh(self):
        raise NotImplementedError


class NotificationConnect:
    """
    Class for sending WS notifications from CARA envoy
    """

    def send_session_expired(self):
        raise NotImplementedError
