import signal
import logging
import sys
from pathlib import Path
from types import FrameType

import re

import subprocess

import os
from typing import Optional

from carta_envoy.connect import RedisConnect

"""
Setup and Launch a CARTA session
"""

CARTA_PROCESS: Optional[subprocess.Popen] = None


class CartaLauncher:
    def __init__(self, settings: dict):
        self.logger = logging.getLogger("carta_envoy")
        self.settings = settings
        self.signal = self.signal_handler
        self.redis_connect = RedisConnect(settings)

    def signal_handler(self, signum: int, frame: FrameType):
        """
        Handle a signal received by tearing down the program and exiting
        NOTE: Parameters signum and frame are convention for signal handlers, even if they don't end up used

        :param signum: Signal number
        :param frame: Stack frame
        """
        if signum == signal.SIGINT:
            self.logger.warning("WARNING: SIGINT received. Exiting.")
        elif signum == signal.SIGALRM:
            self.logger.warning("WARNING: Timeout elapsed. Exiting.")
        self.teardown()
        sys.exit()

    def prepare_for_carta(self):
        ports_and_url = self.redis_connect.prepare_redis()
        signal.signal(signal.SIGINT, self.signal)
        return ports_and_url

    def launch(self):
        self.logger.info("RUNNING CARTA!")
        prepare = self.prepare_for_carta()
        self.run_carta(
            self.settings["carta_path"],
            int(self.settings["timeout"]),
            Path(self.settings["data_location"]),
            prepare["front_end_port"],
            prepare["back_end_port"],
            prepare["carta_url"],
        )

    def run_carta(
        self,
        path_to_carta: str,
        timeout_minutes: int,
        file_browser_path: Path,
        front_end_port: int,
        back_end_port: int,
        carta_url: str,
    ):
        """
        Runs CARTA executable. Will close after a certain amount of time.

        :param path_to_carta: Path to CARTA executable
        :param timeout_minutes: Number of minutes to execute CARTA for before stopping it
        :param file_browser_path: Path to use for CARTA file browser
        :param front_end_port: Port to run the CARTA front end with
        :param back_end_port: Port to run the CARTA back end with
        :param carta_url: URL used to access CARTA from a web browser
        """
        global CARTA_PROCESS

        self.logger.info(f"file_browser_path: {file_browser_path}")
        if not file_browser_path.exists():
            self.logger.error(
                f"ERROR: Invalid file browser path. {file_browser_path}: Directory does not exist."
            )
            return
        elif not file_browser_path.is_dir():
            self.logger.error(
                f"ERROR: Invalid file browser path. {file_browser_path}: Path is not a directory."
            )
            return

        if self.settings["useCarta"]:
            self.logger.info(f"Running CARTA with a timeout of {timeout_minutes} minute(s)...")
            if re.search(r"CARTA.AppImage", path_to_carta):
                # Command uses AppImage, include --remote option
                carta_command = [
                    path_to_carta,
                    "--remote",
                    f"--port={str(back_end_port)}",
                    f"--fport={str(front_end_port)}",
                    f"--folder={file_browser_path}",
                    f"--root={file_browser_path}",
                ]
            else:
                carta_command = [
                    path_to_carta,
                    f"--port={str(back_end_port)}",
                    f"--fport={str(front_end_port)}",
                    f"--folder={file_browser_path}",
                    f"--root={file_browser_path}",
                ]
            try:
                CARTA_PROCESS = subprocess.Popen(
                    carta_command,
                    preexec_fn=os.setpgrp(),
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
            except (OSError, FileNotFoundError, subprocess.SubprocessError) as err:
                self.teardown()
                sys.exit(f"ERROR: Failed to launch CARTA: {err}")
            else:
                # Activate timeout handler
                signal.signal(signal.SIGALRM, self.signal)
                signal.alarm(60 * timeout_minutes)
                self.logger.info(f"To access CARTA, enter this URL into a web browser: {carta_url}")
                self.logger.info("Press Ctrl+C to quit.")

                # Loop until process ends or timeout elapses, whichever happens first
                while CARTA_PROCESS.returncode is None:
                    CARTA_PROCESS.poll()
        else:
            self.logger.info("Running locally...")
            return

    def teardown(self):
        """
        Deletes instance's Redis keys from the Redis server
        """
        self.logger.info("Cleaning up...")
        self.redis_connect.remove_redis_values()

        if CARTA_PROCESS and CARTA_PROCESS.returncode is None:
            # If CARTA is running, kill it
            try:
                # End the process (simulate Ctrl+Z)
                CARTA_PROCESS.send_signal(signal.SIGSTOP)
                # Kill the process
                os.killpg(os.getpgid(CARTA_PROCESS.pid), signal.SIGTERM)
                CARTA_PROCESS.wait()
            except OSError:
                # Process isn't running; just pass
                pass
            else:
                self.logger.info("Killed running CARTA process.")
        else:
            self.logger.warning("WARNING: CARTA not running.")
