"""
Tests for carta_envoy.connect
"""
from unittest.mock import patch

import pytest

from carta_envoy.connect import RedisConnect
from carta_envoy.utilities import generate_random_str

test_settings = {
    "timeout": 1,
    "carta_path": "/fake/path/to/nowhere",
    "redis_host": "noplace.nrao.edu",
    "redis_port": "6397",
    "redis_password": "password",
    "reverse_proxy": "black-abyss.nrao.edu",
    "data_location": ".",
    "useCarta": False,
}

redis_connect = RedisConnect(settings=test_settings)


@pytest.fixture
def redis():
    """
    Pytest fixture that provides a sandbox Redis server for testing purposes

    :return Sandbox Redis server
    """
    import fakeredis

    return fakeredis.FakeStrictRedis()


class TestRedisConnect:
    def test_generate_ids(self):
        pass

    def test_generate_url(self):
        pass

    def test_get_available_port(self):
        """
        Tests that get_available_port returns an available port on localhost
        """
        import socket

        with socket.socket(family=socket.AF_INET, type=socket.SOCK_STREAM) as s:
            try:
                s.bind(("", redis_connect.get_available_port()))
            except OSError as e:
                # Port not available
                print(e)
                assert False

    def test_prepare_redis(self):
        pass

    @pytest.mark.skip("Method not yet implemented")
    def test_send_carta_url_to_rh(self):
        raise NotImplementedError

    def test_connect_to_redis_error(self):
        """
        Tests that a connection to an invalid Redis server correctly throws an exception
        """
        import redis

        with pytest.raises(redis.ConnectionError):
            redis_connect.connect_to_redis()

    def test_get_redis_values(self):
        """
        Tests that filled out Redis values are given correctly
        """
        random_id = generate_random_str()
        front_end_port = 7777
        back_end_port = 6464
        reverse_proxy_host = "test"
        front_end = f"carta-front-{random_id}"
        back_end = f"carta-back-{random_id}"
        values = [
            f"traefik/http/routers/{front_end}/rule",
            f"traefik/http/routers/{front_end}/service",
            f"traefik/http/services/{front_end}/loadbalancer/servers/0/url",
            f"traefik/http/routers/{front_end}/middlewares/0",
            f"traefik/http/routers/{back_end}/rule",
            f"traefik/http/routers/{back_end}/service",
            f"traefik/http/services/{back_end}/loadbalancer/servers/0/url",
            f"traefik/http/routers/{back_end}/middlewares/0",
            f"traefik/http/routers/{back_end}/middlewares/1",
        ]
        assert (
            value
            in redis_connect.get_redis_values(
                reverse_proxy_host, random_id, random_id, front_end_port, back_end_port
            ).keys()
            for value in values
        )

    def test_check_for_duplicate_values(self, redis):
        """
        Tests that duplicate Redis values are correctly reported

        :param redis: Custom fixture that provides a sandbox Redis server
        """
        redis_connect.conn = redis
        redis_connect.generated_ids = {"front_end_id": "1234abcd", "back_end_id": "5678efgh"}
        redis.set("duplicate", "value")
        with patch("carta_envoy.connect.RedisConnect.get_redis_values") as values:
            redis_connect.check_for_duplicate_values({"duplicate": "value"}, 9897, 6543)
            assert values.call_count == 1
            values.call_count = 0
            redis_connect.check_for_duplicate_values({"not_duplicate": "value"}, 9897, 6543)
            assert values.call_count == 0
